/*
 * render.cpp
 *
 *  Created on: 31.01.2009
 *      Author: Лёха
 */

#include <stdlib.h>
#include "Tracer.h"
#include "Render.h"
#include <iostream>
#include <QTime>
#include <QThread>
#include "config.h"

using namespace std;

class TraceThread : public QThread {

	public:
		QImage * img;
		double HalfWidth, HalfHeight, hx, hy;
		int nx, ny;

		int startLine;

	protected:
		void run() {
			Ray ray; // pixel ray
			Vector Color;
			int i, j;
			double x, y;

			for ( i = startLine; i < ny; i += _MULTY_THREAD_COUNT) {
				y = HalfHeight - hy * i;
				x = - HalfWidth;
				for ( j = 0; j < nx; j++ ) {
					x += hx;
					Camera(x, y, ray);
					Color = Trace ( Air, 1.0, ray );
					Clip ( Color );
					img->setPixel(j, i, qRgb((int)round(Color.x * 255),
											(int)round(Color.y * 255),
											(int)round(Color.z * 255)));
				}
			}
		}
};

class TraceThreadAA : public QThread {

	public:
		QImage * img;
		double HalfWidth, HalfHeight, hx, hy;
		int nx, ny, nxSub, nySub;

		int startLine;

	protected:
		void run() {
			Ray ray; // pixel ray
			Vector Color;
			int i, j;
			double x, y, x1, y1;
			int PrimarySamples = nxSub*nySub;
			double hxSub = hx / nxSub;
			double hySub = hy / nySub;

			for ( i = startLine; i < ny; i += _MULTY_THREAD_COUNT) {
				y = HalfHeight - hy * i;
				x = - HalfWidth;
				for ( j = 0; j < nx; j++ ) {
					x += hx;
					x1 = x - 0.5 * hx;
					y1 = y - 0.5 * hy;
					Color = 0;
					for ( int iSub = 0; iSub < nxSub; iSub++ ) {
						for ( int jSub = 0; jSub < nySub; jSub++ ) {
							Camera (x1 + hxSub * (iSub+Rnd()), y1+hySub*(jSub+Rnd()), ray);
							Color += Trace ( Air, 1.0, ray );
						}
					}
					Color /= PrimarySamples;
					Clip(Color);
					img->setPixel(j, i, qRgb((int)round(Color.x * 255),
											 (int)round(Color.y * 255),
											 (int)round(Color.z * 255)));
				}
			}
		}
};

template <typename TraceClassName> bool AllThreadFinish(TraceClassName * threads[]) {
	for(int i = 0; i < _MULTY_THREAD_COUNT; i++) {
		if(threads[i]->isRunning()) {
			return false;
		}
	}
	return true;
}

QImage * _RenderScene ( double HalfWidth, double HalfHeight, int nx, int ny) {
	QImage *img = new QImage(nx, ny, QImage::Format_RGB32);
	double hx = 2.0 * HalfWidth / nx; // pixel width.
	double hy = 2.0 * HalfHeight / ny; // pixel height
	#if _MULTY_THREAD == true
		TraceThread *TraceLine[_MULTY_THREAD_COUNT];

		for(int t = 0; t < _MULTY_THREAD_COUNT; t++) {
			TraceLine[t] = new TraceThread;
			TraceLine[t]->img = img;
			TraceLine[t]->HalfWidth = HalfWidth;
			TraceLine[t]->HalfHeight = HalfHeight;
			TraceLine[t]->hx = hx;
			TraceLine[t]->hy = hy;
			TraceLine[t]->nx = nx;
			TraceLine[t]->ny = ny;
			TraceLine[t]->startLine = t;

			TraceLine[t]->start(_MULTY_THREAD_PRIORITY);
		}
		while(!AllThreadFinish(TraceLine));
	#else
		double x, y; // sample point
		y = HalfHeight;
		Ray ray; // pixel ray
		Vector Color;
		int i, j;
		for ( i = 0; i < ny; i++ ) {
			y -= hy;
			for ( j = 0, x = - HalfWidth; j < nx; j++, x += hx ) {
					Camera(x, y, ray);
					Color = Trace ( Air, 1.0, ray );
					Clip ( Color );
					img->setPixel(j, i, qRgb((int)round(Color.x * 255),
											(int)round(Color.y * 255),
											(int)round(Color.z * 255)));
			}
		}
	#endif
	return img;
};
/*
 * int nxSub
 * int nySub
 * смазывания в px
 * */
QImage * _RenderScene ( double HalfWidth, double HalfHeight, int nx, int ny, int nxSub, int nySub) {
	QImage *img = new QImage(nx, ny, QImage::Format_RGB32);
	double hx = 2.0 * HalfWidth / nx; // pixel width
	double hy = 2.0 * HalfHeight / ny; // pixel height
	#if _MULTY_THREAD == true
		TraceThreadAA *TraceLineAA[_MULTY_THREAD_COUNT];

		for(int t = 0; t < _MULTY_THREAD_COUNT; t++) {
			TraceLineAA[t] = new TraceThreadAA;
			TraceLineAA[t]->img = img;
			TraceLineAA[t]->HalfWidth = HalfWidth;
			TraceLineAA[t]->HalfHeight = HalfHeight;
			TraceLineAA[t]->hx = hx;
			TraceLineAA[t]->hy = hy;
			TraceLineAA[t]->nx = nx;
			TraceLineAA[t]->ny = ny;
			TraceLineAA[t]->startLine = t;
			TraceLineAA[t]->nxSub = nxSub;
			TraceLineAA[t]->nySub = nySub;

			TraceLineAA[t]->start(_MULTY_THREAD_PRIORITY);
		}
		while(!AllThreadFinish(TraceLineAA));
	#else
		double hxSub = hx / nxSub;
		double hySub = hy / nySub;
		double x1, y1;
		double x, y; // sample point
		int i, j;
		Ray ray; // pixel ray
		int PrimarySamples = nxSub*nySub; // of samples for each pixel
		Vector Color;
		for ( i = 0, y = HalfHeight; i < ny; i++, y -= hy ) {
			for ( j = 0, x = - HalfWidth; j < nx; j++, x += hx ) {
				x1 = x - 0.5 * hx;
				y1 = y - 0.5 * hy;
				Color = 0;
				for ( int iSub = 0; iSub < nxSub; iSub++ ) {
					for ( int jSub = 0; jSub < nySub; jSub++ ) {
						Camera (x1 + hxSub * (iSub+Rnd()), y1+hySub*(jSub+Rnd()), ray);
						Color += Trace ( Air, 1.0, ray );
					}
				}
				Color /= PrimarySamples;
				Clip(Color);
				img->setPixel(j, i, qRgb((int)round(Color.x * 255),
										 (int)round(Color.y * 255),
										 (int)round(Color.z * 255)));
			}
		}
	#endif
	return img;
};

/*
 * Variance - критерий точности
 * */
QImage * _RenderScene ( double HalfWidth, double HalfHeight, int nx, int ny, int nxSub, int nySub, double Variance) {
	QImage *img = new QImage(nx, ny, QImage::Format_RGB32);
	double x, y; // sample point
	Ray ray; // pixel ray
	double hx = 2.0 * HalfWidth / nx; // pixel size
	double hy = 2.0 * HalfHeight / ny;
	double hxSub = hx / nxSub;
	double hySub = hy / nySub;
	double Disp; // dispersion squared
	int i, j;
	Vector Color;
	Vector Sum;
	Vector Mean;
	int Count;
	for ( i = 0, y = HalfHeight; i < ny; i++, y -= hy ) {
		for ( j = 0, x = - HalfWidth; j < nx; j++, x += hx ) {
			double x1 = x - 0.5 * hx;
			double y1 = y - 0.5 * hy;
			double d;
			Sum = 0;
			Disp = 0;
			Count = 0;
			do {
				for ( int iSub = 0; iSub < nxSub; iSub++ ) {
					for ( int jSub = 0; jSub < nySub; jSub++ ) {
						Camera (x1 + hxSub * (iSub + Rnd()), y1 + hySub * (jSub+Rnd()), ray);
						Color = Trace ( Air, 1.0, ray );
						Sum += Color;
						Disp += Color & Color;
						Count++;
					}
				}
				Mean = Sum / Count;
				d = (Disp / Count - (Mean & Mean)) * Count / (Count - 1);
			} while ( d / Count >= Variance && Count < MaxLevel );
			Clip(Mean);
			img->setPixel(j, i, qRgb((int)round(Mean.x * 255), (int)round(Mean.y * 255), (int)round(Mean.z * 255)));
		}
	}
	return img;
};

QImage * RenderScene ( double HalfWidth, double HalfHeight, int nx, int ny) {
	QTime t;
	QImage * img;
	TotalRays = 0l;
	Level = 0;
	#if _TRACE_DEBUG == true
		t.start();
		cout << "Start: " << t.currentTime().toString("hh:mm:ss.zzz").toStdString();
	#endif
	if(_TRACE_TYPE == 0) {
		img = _RenderScene ( HalfWidth, HalfHeight, nx, ny);
	} else if(_TRACE_TYPE == 1) {
		img = _RenderScene ( HalfWidth, HalfHeight, nx, ny, _ANTI_ALIAS_X, _ANTI_ALIAS_Y);
	} else if(_TRACE_TYPE == 2) {
		img = _RenderScene ( HalfWidth, HalfHeight, nx, ny, _ANTI_ALIAS_X, _ANTI_ALIAS_Y, _ACCURACY);
	} else {
		//type 1
		img = _RenderScene ( HalfWidth, HalfHeight, nx, ny, _ANTI_ALIAS_X, _ANTI_ALIAS_Y);
	}
	#if _TRACE_DEBUG == true
		cout << ", Stop: " <<  t.currentTime().toString("hh:mm:ss.zzz").toStdString() <<
			    ", Time: " << t.elapsed() <<
				", Rays: " << TotalRays <<
				", Lights: " << Scene->LightsCount <<
				", Solids: " << Scene->SolidsCount <<
				", Trace Type: " << _TRACE_TYPE;
		if(_MULTY_THREAD) {
			cout << ", Threads: " << _MULTY_THREAD_COUNT <<
					", Priority: " << _MULTY_THREAD_PRIORITY;
		}
		cout << endl;
	#endif

	return img;
};
