/*
 * Tracer.cpp
 *
 *  Created on: 15.11.2008
 *      Author: Лёха
 */

#include "Tracer.h"

///////////////////// Globals ////////////////////
Vector Eye (0, 0, 0 ); // camera position
Vector EyeDir (0, 0, 1 ); // viewing направление
Vector Vx ( 1, 0, 0 ); // базовая поверхность изображения
Vector Vy ( 0, 1, 0 );
Vector Ambient ( 1.0 ); // ambient light intensity
Vector Background ( 0.0, 0.05, 0.05 ); // background
Medium Air = { 1, 0 }; // basic mediums : Air
Medium Glass = { 1.5, 0 }; // glass
int Level = 0; // current recusion level
double Threshold = _THRESHOLD; // accuracy of computations
int MaxLevel = _MAX_LEVEL_RECURSION; //max. levels of recursion
Environment * Scene;
long TotalRays = 0l;
int threadCount = _MULTY_THREAD_COUNT;

#if (_MULTY_THREAD && _MULTY_THREAD_LOCK && threadCount > 1) == true
	#include <QMutex>
	QMutex lock;
#endif

GObject :: ~GObject () {
	if ( Mapping != NULL ) delete Mapping;
	for ( Texture * m = Material; m != NULL; m = Material ) {
		Material = Material->next;
		delete m;
	}
};

void GObject :: Add ( Texture * m ) {
	m->next = Material;
	m->object = this;
	Material = m;
};

void GObject :: Add ( Perturbation * p ) {
	p->object = this;
	pertr = p;
};

void GObject :: FindTexture ( Vector& p, SurfaceData& t ) {
	t = DefMaterial;
	t.n = FindNormal ( p );
	if ( Mapping != NULL ) t.MapCoord = Mapping -> Apply ( p );
	for ( Texture * m = Material; m != NULL; m = m->next )
		m -> Apply ( p, t );
};

void GObject :: FindPerturbation(Vector& view) {
	if(pertr != NULL) pertr->Apply(view);
};

//////////////////// Environment methods //////////////////////
Environment :: ~Environment () {
//delete all contained objects
	for ( int i = 0; i < LightsCount; i++ )
		delete Light [i];
	for ( int i = 0; i < SolidsCount; i++ )
		delete Solid [i];
};

void Environment :: Add ( LightSource * l ) {
	if ( LightsCount < _MAX_LIGHTS - 1 ) Light [LightsCount++] = l;
};

void Environment :: Add ( GObject * o ) {
	if ( SolidsCount < _MAX_SOLIDS - 1 ) Solid [SolidsCount++] = o;
};

// find closest intersection with scene objects
GObject * Environment :: Intersect ( Ray& ray, double& t ) {
	GObject * ClosestObj = NULL;
	double ClosestDist = _INFINITY;
	for ( int i = 0; i < SolidsCount; i++ ) { // check every object
		Solid [i]->FindPerturbation(ray.Dir);
		if ( Solid [i]->Intersect ( ray, t ) )
			if ( t < ClosestDist ) {
				ClosestDist = t;
				ClosestObj = Solid [i];
			}
	}
	t = ClosestDist;
	return ClosestObj;
};

Vector Environment :: ShadeBackground ( Ray& ray ) {
	return Background;
};

////////////////////////// Functions ////////////////////
void SetCamera ( Vector& Org, Vector& Dir, Vector& UpDir ) {
	Eye = Org; // eye point
	EyeDir = Dir; // viewing direction
	Vector tmp;
	tmp = UpDir ^ Dir;
	Vx = Normalize ( tmp );
	tmp = Dir ^ Vx;
	Vy = Normalize ( tmp );
};

// get a pixel ray for a given screen point ( x, у )
void Camera ( double x, double y, Ray& ray ) {
	ray.Org = Eye;
	Vector tmp = EyeDir + Vx * x + Vy * y;
	ray.Dir = Normalize (tmp);
};

// Trace a given ray through the scene
Vector Trace ( Medium& CurMed, double Weight, Ray& ray ) {
	GObject * Obj;
	double t = _INFINITY;
	Vector Color;
	Vector tmp;
	#if (_MULTY_THREAD && _MULTY_THREAD_LOCK && threadCount > 1) == true
		lock.lock();
		Level++;
		TotalRays++;
		lock.unlock();
	#else
		Level++;
		TotalRays++;
	#endif
	if ( ( Obj = Scene->Intersect ( ray, t ) ) != NULL ) {
		tmp = ray.Point(t);
		Color = Shade ( CurMed, Weight, tmp, ray.Dir, Obj );
		if (CurMed.Betta > Threshold)
			Color *= exp (-t * CurMed.Betta);
	}
	else
		Color = Scene->ShadeBackground ( ray );
	#if (_MULTY_THREAD && _MULTY_THREAD_LOCK && threadCount > 1) == true
		lock.lock();
		Level--;
		lock.unlock();
	#else
		Level--;
	#endif
	return Color;
}

// compute light coming from point p in the direction View
Vector Shade ( Medium& CurMed, double Weight, Vector& p, Vector& View, GObject * Obj ) {
	SurfaceData txt;
	Ray ray;
	Vector Color;
	Vector l; // light vector
	double Sh; // light shadow coeff.
	Vector h; // vector between -View arc light
	double In, vn;
	int Entering = 1; // flag whether we're entering
	Obj->FindTexture ( p, txt );
	if ( ( vn = View & txt.n ) > 0 ) { // force ( -View, n ) > 0
		txt.n = -txt.n;
		vn = -vn;
		Entering = 0;
	}
	ray.Org = p;
	Color = Ambient * txt.Color * txt.Ka; // get ambient light (фон)
	Vector tmp;
	for ( int i = 0; i < Scene->LightsCount; i++ ) {
		if (( Sh = Scene->Light[i]->Shadow( p, l ) ) > Threshold ) {
			if (( In = l & txt.n ) > Threshold ) {// light is visible
				if ( txt.Kd > Threshold ) {
					Color += Scene->Light[i] -> Color * txt.Color*(txt.Kd * Sh * In);
				}
				if ( txt.Ks > Threshold ) {
					tmp = l - View;
					h = Normalize (tmp);
					Color += Scene->Light[i]->Color * (txt.Ks * Sh * pow (txt.n&h, txt.p));
				}
			}
		}
	}
	double rWeight = Weight * txt.Kr; // weight of reflected ray //отраженный
	double tWeight = Weight * txt.Kt; // weight of transmitted //преломленный
	// check for reflected ray
	if ( rWeight > Threshold && Level < MaxLevel ) {
		ray.Dir = View - txt.n * ( 2 * vn ); // get reflected ray //отраж
		Color += txt.Kr * Trace ( CurMed, rWeight, ray );
	}
	// check for transmitted
	if (tWeight > Threshold && Level < MaxLevel) {
		double Eta = CurMed.nRefr/(Entering?txt.Med.nRefr:Air.nRefr);
		double ci = - vn; // cosine of incident angle
		double ctSq = 1 + Eta*Eta*( ci*ci - 1 );
		if ( ctSq > Threshold ) { // not a Total Internal Reflection
			ray.Dir = View * Eta + txt.n * ( Eta*ci - sqrt (ctSq) );
			if ( Entering ) { // ray enters object
				Color += txt.Kr * Trace ( txt.Med, tWeight, ray );
			} else { // ray leaves object
				Color += txt.Kr * Trace ( Air, tWeight, ray );
			}
		}
	}
	return Color;
};

int BoundingVolume :: Check ( Ray& r) {
	return true; //TODO
};

GObject * BoundingVolume :: Intersect ( Ray& ray, double& t ) {
	if ( !Check ( ray ) ) return NULL;
	GObject * Found = NULL;
	GObject * tmp;
	BoundingVolume * vol = child;
	double t1;
	for ( t = _INFINITY; vol != NULL; vol = vol -> next ) {
		if ( ( tmp = vol -> Intersect ( ray, t1 ) ) != NULL ) {
			if ( t1 < t ) {
				Found = tmp;
				t = t1;
			}
		}
	}
	if ( Obj != NULL ) {
		if ( Obj -> Intersect ( ray, t1 ) ) {
			if ( t1 < t ) {
				Found = Obj;
				t = t1;
			}
		}
	}
	return Found;
};

