/*
 * Tracer.h
 *
 *  Created on: 15.11.2008
 *      Author: Лёха
 */

#ifndef TRACER_H_
#define TRACER_H_

// File Tracer.h
#include <stdlib.h>
#include "Vector.h"
#include "config.h"

class GObject;
class Texture;

struct Medium { // main properties of the medium
	double nRefr; // refraction coefficient
	double Betta; // attenuation coefficient
};

struct SurfaceData { // surface characteristics at a given point
	double Ka; // ambient light coefficient //фон
	double Kd; // diffuse light coefficient //диф отраж
	double Ks; // specular light coefficient //зеркальный
	double Kr; // reflected ray coefficient //отраженный
	double Kt; // transparent light coefficient //преломленный
	Vector Color; // object's color
	Medium Med; // medium of the object
	int p; // Phong's coeff.
	Vector n; // normal at a given point
	Vector MapCoord; //mapping coordinates
};

class LightSource { // model of an abstract light source
	public:
		Vector Color;
		LightSource () { Color = 1; };
		virtual ~LightSource () {}; // force virtual destructor
		virtual double Shadow ( Vector&, Vector& ) = 0;
};

class Texture {// generic texture class
	public:
		Texture * next;
		GObject * object;
		Vector Offs;
		Vector Scale;
		Texture () {
			next = NULL;
			object = NULL;
			Offs = 0;
			Scale = 1;
		};
		virtual ~Texture () {};
		virtual void Apply ( Vector& p, SurfaceData& t ) = 0;
};

class Perturbation {// generic perturbation class
	public:
		GObject * object;
		Perturbation () {
			object = NULL;
		};
		virtual ~Perturbation () {};
		virtual void Apply (Vector& view) = 0;
};

class Map { // generic mapping class
	public:
		virtual ~Map () {}; // force virtual destructor
		virtual Vector Apply ( Vector& ) = 0; // map point
		virtual void FindTangent ( Vector&, Vector&, Vector& ) = 0;
};

class GObject {// model of an abstract geometric object
	public:
		SurfaceData DefMaterial; // default material
		Map * Mapping;
		Texture * Material;
		Perturbation * pertr;
		GObject () {
			Mapping = NULL;
			Material = NULL;
			pertr = NULL;
		};
		virtual ~GObject (); // force virtual destructor
		void FindTexture ( Vector& p, SurfaceData& t );
		void FindPerturbation ( Vector& view );
		void Add (Texture *);
		void Add (Perturbation *);
		virtual int Intersect ( Ray&, double& ) = 0;
		virtual Vector FindNormal ( Vector& ) = 0;
};

class Environment { // simplest model of environment{
	public:
		LightSource * Light [_MAX_LIGHTS];
		int LightsCount;
		GObject * Solid [_MAX_SOLIDS];
		int SolidsCount;
		Environment () { LightsCount = SolidsCount =0; };
		virtual ~Environment ();
		void Add ( LightSource * );
		void Add ( GObject * );
		virtual GObject * Intersect ( Ray&, double& );
		virtual Vector ShadeBackground ( Ray& );
};

// Bounding Volume Hierarchy
class BoundingVolume {
	public:
		BoundingVolume * child;
		BoundingVolume * next;
		GObject * Obj;
		virtual int Check ( Ray& ) = 0;
		GObject * Intersect ( Ray&, double& );
};

/////////////////////// Globals //////////////////////////////
extern Vector Eye; // camera position
extern Vector EyeDir; // camera viewing direction
extern Vector Vx, Vy; // image plane basis (Vx-hor, Vy-vert)
extern Medium Air; // basic materials
extern Medium Glass;
extern int Level; // current recusion level
extern double Threshold;
extern int MaxLevel; // max. levels of recursion
extern Vector Ambient; // ambient light intensity
extern Vector Background;
extern Environment * Scene;
extern long TotalRays;

////////////////////// Function definitions /////////////////////
void Camera ( double, double, Ray& ); // get ray
void SetCamera ( Vector&, Vector&, Vector&); // set new camera
Vector Trace ( Medium&, double, Ray& ); // trace a ray
Vector Shade ( Medium&, double, Vector&, Vector&, GObject * );
inline double SawWave ( double ); //TODO
inline double SineWave ( double x ) {
	return 0.5 - ( 1.0 + sin ( x ) );
}
inline double Mod ( double x, double y ) {
	if ( ( x = fmod ( x, y ) ) < 0 ) return x + y;
		else return x;
}
inline double Rnd () {
	return ( (double) rand() ) / (double) RAND_MAX;
}

#endif /* TRACER_H_ */
